from .core import UnitedStates
from ..registry_tools import iso_register


# FIXME: According to wikipedia, Kansas only has all federal holidays, except
# the Columbus Day and Washington's Birthday.
# Unfortunately, other sources mention XMas Eve for 2018, but not for other
# years.
# I'm a bit sad here...
#
# Sources to lookup, if you want to help:
# * http://www.admin.ks.gov/docs/default-source/ops/holidays/holidays2018.pdf
# * http://www.kansas.gov/employee/documents/2017calendar.pdf
# * https://publicholidays.us/kansas/2018-dates/
# * https://en.wikipedia.org/wiki/Public_holidays_in_the_United_States#Kansas
# * https://publicholidays.us/kansas/2018-dates/


@iso_register('US-KS')
class Kansas(UnitedStates):
    """Kansas"""
    include_federal_presidents_day = False
    include_columbus_day = False



"""
Virginia module

You may or may want not to treat the Day before Thanksgiving as a non-working
day by implementing the following class:

.. code::

    from workalenda.usa import Virginia as VirginiaBase

    class Virginia(VirginiaBase):
        include_thanksgiving_wednesday = False

"""
from ..core import WED, FRI
from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-VA')
class Virginia(UnitedStates):
    """Virginia"""
    include_christmas_eve = True
    include_thanksgiving_friday = True
    include_boxing_day = True
    presidents_day_label = "George Washington Day"
    # Virginia specific. By default, it's treated as a holiday, but
    # you may chose to exclude it for you own uses. See the module doc.
    include_thanksgiving_wednesday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(
            (self.get_nth_weekday_in_month(year, 1, FRI, 3),
             "Lee-Jackson Day")
        )
        if self.include_thanksgiving_wednesday:
            days.append(
                (self.get_nth_weekday_in_month(year, 11, WED, 4),
                 "Day before Thanksgiving (start at noon)")
            )

        return days



from datetime import date

from ..core import MON, WesternCalendar
from ..registry_tools import iso_register


@iso_register('GG')
class Guernsey(WesternCalendar):
    'Guernsey'

    include_easter_monday = True
    include_boxing_day = True
    shift_new_years_day = True
    include_good_friday = True

    def get_spring_bank_holiday(self, year):
        spring_bank_holiday = Guernsey \
            .get_last_weekday_in_month(year, 5, MON)
        return (
            spring_bank_holiday,
            "Spring Bank Holiday"
        )

    def get_early_may_bank_holiday(self, year):
        """
        Return Early May bank holiday
        """
        # Special case in 2020, for the 75th anniversary of the end of WWII.
        if year == 2020:
            return (
                date(year, 5, 8),
                "Early May bank holiday (VE day)"
            )

        return (
            Guernsey.get_nth_weekday_in_month(year, 5, MON),
            "Early May Bank Holiday"
        )

    def get_summer_bank_holiday(self, year):
        return (
            Guernsey.get_last_weekday_in_month(year, 8, MON),
            "Summer Bank Holiday"
        )

    def get_liberation_day(self, year):
        return (date(year, 5, 9), "Liberation Day")

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(self.get_early_may_bank_holiday(year))
        days.append(self.get_spring_bank_holiday(year))
        days.append(self.get_summer_bank_holiday(year))
        days.append(self.get_liberation_day(year))
        return days



from unittest import TestCase
from ..usa import (
    UnitedStates,
    Alabama,
    Alaska,
    Arizona,
    Arkansas,
    California,
    Colorado,
    Connecticut,
    Delaware,
    DistrictOfColumbia,
    Florida,
    Georgia,
    Hawaii,
    Idaho,
    Illinois,
    Indiana,
    Iowa,
    Kansas,
    Kentucky,
    Louisiana,
    Maine,
    Maryland,
    Massachusetts,
    Michigan,
    Minnesota,
    Mississippi,
    Missouri,
    Montana,
    Nebraska,
    Nevada,
    NewHampshire,
    NewJersey,
    NewMexico,
    NewYork,
    NorthCarolina,
    NorthDakota,
    Ohio,
    Oklahoma,
    Oregon,
    Pennsylvania,
    RhodeIsland,
    SouthCarolina,
    SouthDakota,
    Tennessee,
    Texas,
    Utah,
    Vermont,
    Virginia,
    Washington,
    WestVirginia,
    Wisconsin,
    Wyoming
)

from ..registry import registry


class RegistryUsa(TestCase):

    def _check_all_states(self, classes):
        self.assertIn(Alabama, classes)
        self.assertIn(Alaska, classes)
        self.assertIn(Arizona, classes)
        self.assertIn(Arkansas, classes)
        self.assertIn(California, classes)
        self.assertIn(Colorado, classes)
        self.assertIn(Connecticut, classes)
        self.assertIn(Delaware, classes)
        self.assertIn(DistrictOfColumbia, classes)
        self.assertIn(Florida, classes)
        self.assertIn(Georgia, classes)
        self.assertIn(Hawaii, classes)
        self.assertIn(Idaho, classes)
        self.assertIn(Illinois, classes)
        self.assertIn(Indiana, classes)
        self.assertIn(Iowa, classes)
        self.assertIn(Kansas, classes)
        self.assertIn(Kentucky, classes)
        self.assertIn(Louisiana, classes)
        self.assertIn(Maine, classes)
        self.assertIn(Maryland, classes)
        self.assertIn(Massachusetts, classes)
        self.assertIn(Michigan, classes)
        self.assertIn(Minnesota, classes)
        self.assertIn(Mississippi, classes)
        self.assertIn(Missouri, classes)
        self.assertIn(Montana, classes)
        self.assertIn(Nebraska, classes)
        self.assertIn(Nevada, classes)
        self.assertIn(NewHampshire, classes)
        self.assertIn(NewJersey, classes)
        self.assertIn(NewMexico, classes)
        self.assertIn(NewYork, classes)
        self.assertIn(NorthCarolina, classes)
        self.assertIn(NorthDakota, classes)
        self.assertIn(Ohio, classes)
        self.assertIn(Oklahoma, classes)
        self.assertIn(Oregon, classes)
        self.assertIn(Pennsylvania, classes)
        self.assertIn(RhodeIsland, classes)
        self.assertIn(SouthCarolina, classes)
        self.assertIn(SouthDakota, classes)
        self.assertIn(Tennessee, classes)
        self.assertIn(Texas, classes)
        self.assertIn(Utah, classes)
        self.assertIn(Vermont, classes)
        self.assertIn(Virginia, classes)
        self.assertIn(Washington, classes)
        self.assertIn(WestVirginia, classes)
        self.assertIn(Wisconsin, classes)
        self.assertIn(Wyoming, classes)

    def test_usa_world(self):
        classes = set(registry.region_registry.values())
        self._check_all_states(classes)
        # On top of it, the core class
        self.assertIn(UnitedStates, classes)

    def test_usa_subregion(self):
        # Get all the subregions
        classes = set(registry.get_subregions('US').values())
        self._check_all_states(classes)



from datetime import date

from ..core import (
    IslamicMixin, ChineseNewYearCalendar,
    SAT, SUN
)
from ..registry_tools import iso_register


@iso_register('MY')
class Malaysia(IslamicMixin, ChineseNewYearCalendar):
    "Malaysia"
    # Civil holidays
    include_labour_day = True
    labour_day_label = "Workers' Day"

    # Islamic holidays
    include_nuzul_al_quran = True
    include_eid_al_fitr = True
    length_eid_al_fitr = 2
    eid_al_fitr_label = "Hari Raya Puasa"
    include_day_of_sacrifice = True
    day_of_sacrifice_label = "Hari Raya Haji"
    include_islamic_new_year = True
    include_prophet_birthday = True

    # Most of th Malaysian territory uses these Week-end days
    WEEKEND_DAYS = (SAT, SUN)
    # TODO: Add calendar exceptions

    FIXED_HOLIDAYS = ChineseNewYearCalendar.FIXED_HOLIDAYS + (
        (2, 1, "Federal Territory Day"),
        (8, 31, "National Day"),
        (9, 16, "Malaysia Day"),
        (12, 25, "Christmas Day"),
    )

    # Ref: https://publicholidays.com.my/deepavali/
    MSIA_DEEPAVALI = {
        2010: date(2010, 11, 5),
        2011: date(2011, 10, 26),
        2012: date(2012, 11, 13),
        2013: date(2013, 11, 2),
        2014: date(2014, 10, 22),
        2015: date(2015, 11, 10),
        2016: date(2016, 10, 29),
        2017: date(2017, 10, 18),
        2018: date(2018, 11, 6),
        2019: date(2019, 10, 27),
        2020: date(2020, 11, 14),
        2021: date(2021, 11, 4),
        2022: date(2022, 10, 24),
        2023: date(2023, 11, 13),  # in lieu of 12th Nov
        2024: date(2024, 10, 31),
    }

    # Ref: https://publicholidays.com.my/thaipusam/
    MSIA_THAIPUSAM = {
        2010: date(2010, 1, 30),
        2011: date(2011, 1, 20),
        2012: date(2012, 2, 7),
        2013: date(2013, 1, 28),
        2014: date(2014, 1, 17),
        2015: date(2015, 2, 3),
        2016: date(2016, 1, 25),
        2017: date(2017, 2, 9),
        2018: date(2018, 1, 31),
        2019: date(2019, 1, 21),
        2020: date(2020, 2, 8),
        2021: date(2021, 1, 28),
        2022: date(2022, 1, 18),
        2023: date(2023, 2, 4),
        2024: date(2024, 1, 25),
    }
    chinese_new_year_label = "First Day of Lunar New Year"
    include_chinese_second_day = True
    chinese_second_day_label = "Second Day of Lunar New Year"
    shift_sunday_holidays = True

    def get_variable_days(self, year):
        """
        Malaysia variable days
        """
        days = super().get_variable_days(year)

        # Vesak Day
        days.append(
            (ChineseNewYearCalendar.lunar(year, 4, 15), "Vesak Day"),
        )

        # Add in Deepavali and Thaipusam (hardcoded dates, so no need to shift)
        msia_deepavali = self.MSIA_DEEPAVALI.get(year)
        if not msia_deepavali:
            mdmsg = f'Missing date for Malaysia Deepavali for year: {year}'
            raise KeyError(mdmsg)
        days.append((msia_deepavali, 'Deepavali'))

        msia_thaipusam = self.MSIA_THAIPUSAM.get(year)
        if not msia_thaipusam:
            mtmsg = f'Missing date for Malaysia Thaipusam for year: {year}'
            raise KeyError(mtmsg)
        days.append((msia_thaipusam, 'Thaipusam'))
        return days



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-MS')
class Mississippi(UnitedStates):
    """Mississippi"""
    include_thanksgiving_friday = True
    include_confederation_day = True
    include_columbus_day = False

    martin_luther_king_label = ("Martin Luther King's"
                                " and Robert E. Lee's Birthdays")
    veterans_day_label = "Armistice Day (Veterans Day)"
    national_memorial_day_label = ("National Memorial Day / "
                                   "Jefferson Davis Birthday")



from copy import copy
from datetime import date, timedelta

from ..core import MON, SUN, WesternCalendar
from ..registry_tools import iso_register


@iso_register("BB")
class Barbados(WesternCalendar):
    "Barbados"
    # Civil holidays
    include_labour_day = True
    # Christian holidays
    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_whit_monday = True
    non_computable_holiday_dict = {
        2016: [(date(2016, 12, 27), "Public Holiday")],
        2021: [
            (date(2021, 1, 4), "Public Holiday"),
            (date(2021, 1, 5), "Public Holiday"),
            ],
    }

    # All holiday are shifted if on a Sunday
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (1, 21, "Errol Barrow Day"),
        (4, 28, "National Heroes Day"),
        (11, 30, "Independance Day"),
        (12, 26, "Boxing Day"),  # Not the same as UK Boxing Day
    )

    def get_kadooment_day(self, year):
        """
        First Monday of August.
        """
        return (Barbados.get_nth_weekday_in_month(year, 8, MON),
                "Kadooment Day")

    def get_emancipation_day(self, year):
        emancipation_day = date(year, 8, 1)
        if emancipation_day.weekday() == SUN:
            emancipation_day += timedelta(days=1)
        kadooment_day = self.get_kadooment_day(year)
        if emancipation_day == kadooment_day[0]:
            emancipation_day += timedelta(days=1)
        return (emancipation_day, "Emancipation Day")

    def get_variable_days(self, year):
        """
        Return variable holidays of the Barbados calendar.
        """
        days = super().get_variable_days(year)
        days.append(self.get_kadooment_day(year))
        days.append(self.get_emancipation_day(year))
        non_computable = self.non_computable_holiday(year)
        if non_computable:
            days.extend(non_computable)
        return days

    def non_computable_holiday(self, year):
        non_computable = self.non_computable_holiday_dict.get(year, None)
        return non_computable

    def get_fixed_holidays(self, year):
        """
        Return fixed holidays of the Barbados calendar.

        A shift day is appended if a fixed holiday happens on SUN.
        """
        days = super().get_fixed_holidays(year)
        days_to_shift = copy(days)
        for day, label in days_to_shift:
            if day.weekday() == SUN:
                days.append((day + timedelta(days=1), f"{label} (shifted)"))
        return days



from datetime import date

from ..core import IslamoWesternCalendar, SAT, SUN
from ..registry_tools import iso_register


@iso_register("NG")
class Nigeria(IslamoWesternCalendar):
    "Nigeria"

    # Civil holidays
    include_labour_day = True
    labour_day_label = "Workers' Day"

    # Christian holidays
    include_good_friday = True
    include_easter_monday = True
    include_boxing_day = True

    # Islamic holidays
    include_eid_al_fitr = True
    include_day_of_sacrifice = True

    shift_sunday_holidays = True
    shift_new_years_day = True

    # Explicitly assign these WE days, Nigeria has adopted the western workweek
    WEEKEND_DAYS = (SAT, SUN)

    FIXED_HOLIDAYS = IslamoWesternCalendar.FIXED_HOLIDAYS + (
        (10, 1, "Independence Day"),
    )

    def get_fixed_holidays(self, year):
        """Get fixed holidays.

        Args:
            year (int): Year

        Returns:
            tuple: Tuple of date and label
        """
        days = super().get_fixed_holidays(year)

        democracy_day_label = "Democracy Day"

        if 2000 <= year < 2018:
            days.append((date(year, 5, 29), democracy_day_label))

        if year >= 2018:
            days.append((date(year, 6, 12), democracy_day_label))

        return days



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-ND')
class NorthDakota(UnitedStates):
    """North Dakota"""
    include_columbus_day = False
    include_good_friday = True

# NOTE: At the time of writing, the Public Holidays for USA wikipedia page
# doesn't mention Good Friday, although the official Secretary of State page
# does: http://sos.nd.gov/about-office/holiday-office-closings



from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('AO')
class Angola(WesternCalendar):
    "Angola"
    # Civil holiday
    include_labour_day = True
    labour_day_label = "Dia Internacional do Trabalhador"
    # Christian holidays
    include_fat_tuesday = True
    fat_tuesday_label = "Dia de Carnaval"
    include_good_friday = True
    include_easter_sunday = True
    include_christmas = True
    include_all_souls = True

    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (2, 4, "Dia do Inicio da Luta Armada"),
        (3, 8, "Dia Internacional da Mulher"),
        (4, 4, "Dia da Paz"),
        (9, 17, "Dia do Fundador da Nação e do Herói Nacional"),
        (11, 11, "Dia da Independência Nacional"),
    )



from datetime import date

from ..core import OrthodoxCalendar, MON, daterange, cleaned_date
from ..registry_tools import iso_register


@iso_register('RU')
class Russia(OrthodoxCalendar):
    'Russia'
    # Civil holidays
    include_labour_day = True

    FIXED_HOLIDAYS = OrthodoxCalendar.FIXED_HOLIDAYS + (
        (1, 2, "Day After New Year"),
        (2, 23, "Defendence of the Fatherland"),
        (3, 8, "International Women's Day"),
        (5, 9, "Victory Day"),
        (6, 12, "National Day"),
        (11, 4, "Day of Unity"),
    )

    # Christian holidays
    include_christmas = False

    covid19_2020_start = date(2020, 3, 28)
    covid19_2020_end = date(2020, 4, 30)

    def get_fixed_holidays(self, year):

        if year >= 1992:
            self.labour_day_label = "The Day of Spring and Labour"
        else:
            self.labour_day_label = "International Workers' Day"

        days = super().get_fixed_holidays(year)

        if year >= 2005:
            days.extend([
                (date(year, 1, 3), "Third Day after New Year"),
                (date(year, 1, 4), "Fourth Day after New Year"),
                (date(year, 1, 5), "Fifth Day after New Year"),
                (date(year, 1, 6), "Sixth Day after New Year"),
                (date(year, 1, 8), "Eighth Day after New Year"),
            ])

        if year == 2020:
            index = 1
            for day in daterange(self.covid19_2020_start,
                                 self.covid19_2020_end):
                days.append(
                    (day, f"Non-Working Day (COVID-19) #{index}")
                )
                index += 1

            # Adding May extra days
            days.extend([
                (date(year, 5, 4), "May 4th, 2020 holiday"),
                (date(year, 5, 5), "Day of Spring and Labor"),
            ])

            # Extra COVID-19 in May
            days.extend([
                (date(year, 5, 6), "Covid-19 May #1"),
                (date(year, 5, 7), "Covid-19 May #2"),
                (date(year, 5, 8), "Covid-19 May #3"),
            ])

            # Constitution Vote Public Holiday
            days.append((date(year, 7, 1), "Constitution Vote Public Holiday"))
        elif year == 2021:
            days.extend([
                (date(year, 2, 22), "Day Before Defendence of the Fatherland"),
                (date(year, 11, 5), 'Day After Day of Unity'),
                (date(year, 12, 31), "New Year's Eve"),
            ])
        return days

    def get_calendar_holidays(self, year):
        holidays = super().get_calendar_holidays(year)
        shifts = []
        for day, label in holidays:
            if day.month == 1 and day.day in range(1, 9):
                continue
            # Add an exception for 2020 non-working days due to COVID-19
            if self.covid19_2020_start <= day <= self.covid19_2020_end:
                continue  # pragma: no cover
            if day.weekday() in self.get_weekend_days():
                shifts.append((
                    self.get_first_weekday_after(day, MON),
                    label + " shift"
                ))
        holidays.extend(shifts)
        return holidays

    def is_working_day(self, day,
                       extra_working_days=None, extra_holidays=None):
        day = cleaned_date(day)
        if day == date(2021, 2, 20):
            return True
        return super().is_working_day(
            day,
            extra_working_days=extra_working_days,
            extra_holidays=extra_holidays
        )



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-MD')
class Maryland(UnitedStates):
    """Maryland"""
    thanksgiving_friday_label = "Native American Heritage Day"
    include_thanksgiving_friday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        if Maryland.is_presidential_year(year):
            days.append(self.get_election_day(year))
        return days



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-UT')
class Utah(UnitedStates):
    """Utah"""
    FIXED_HOLIDAYS = UnitedStates.FIXED_HOLIDAYS + (
        (7, 24, "Pioneer Day"),
    )



from datetime import date

from . import GenericCalendarTest
from ..africa.mozambique import Mozambique


class MozambiqueTest(GenericCalendarTest):
    cal_class = Mozambique

    def test_year_new_year_shift(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 1), holidays)
        self.assertNotIn(date(2019, 1, 2), holidays)
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)
        self.assertNotIn(date(2020, 1, 2), holidays)

    def test_n_holidays(self):
        n_holidays = len(self.cal.holidays_set(2019))
        for holiday in self.cal.get_calendar_holidays(2020):
            print(holiday)
        assert n_holidays == 10

    def test_year_2018(self):
        holidays = self.cal.holidays_set(2018)
        # Fixed days section:
        # 1. New Year's Day
        self.assertIn(date(2018, 1, 1), holidays)
        # 2. Mozambican Heroes' Day
        self.assertIn(date(2018, 2, 3), holidays)
        # 3. Mozambican Women's Day
        self.assertIn(date(2018, 4, 7), holidays)
        # 4. Good Friday
        self.assertIn(date(2018, 3, 30), holidays)
        # 5. Labour Day
        self.assertIn(date(2018, 5, 1), holidays)
        # 6. Independence Day
        self.assertIn(date(2018, 6, 25), holidays)
        # 7. Victory Day
        self.assertIn(date(2018, 9, 7), holidays)
        # 8. Armed Forces Day
        self.assertIn(date(2018, 9, 25), holidays)
        # 9. Peace And Reconciliation Day
        self.assertIn(date(2018, 10, 4), holidays)
        # 10. Christmas day
        self.assertIn(date(2018, 12, 25), holidays)

    def test_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        # Fixed days section:
        # 1. New Year's Day
        self.assertIn(date(2019, 1, 1), holidays)
        # 2. Mozambican Heroes' Day
        self.assertIn(date(2019, 2, 3), holidays)
        # 3. Mozambican Women's Day
        self.assertIn(date(2019, 4, 7), holidays)
        # 4. Good Friday
        self.assertIn(date(2019, 4, 19), holidays)
        # 5. Labour Day
        self.assertIn(date(2019, 5, 1), holidays)
        # 6. Independence Day
        self.assertIn(date(2019, 6, 25), holidays)
        # 7. Victory Day
        self.assertIn(date(2019, 9, 7), holidays)
        # 8. Armed Forces Day
        self.assertIn(date(2019, 9, 25), holidays)
        # 9. Peace And Reconciliation Day
        self.assertIn(date(2019, 10, 4), holidays)
        # 10. Christmas day
        self.assertIn(date(2019, 12, 25), holidays)

    def test_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        # Fixed days section:
        # 1. New Year's Day
        self.assertIn(date(2020, 1, 1), holidays)
        # 2. Mozambican Heroes' Day
        self.assertIn(date(2020, 2, 3), holidays)
        # 3. Mozambican Women's Day
        self.assertIn(date(2020, 4, 7), holidays)
        # 4. Good Friday
        self.assertIn(date(2020, 4, 10), holidays)
        # 5. Labour Day
        self.assertIn(date(2020, 5, 1), holidays)
        # 6. Independence Day
        self.assertIn(date(2020, 6, 25), holidays)
        # 7. Victory Day
        self.assertIn(date(2020, 9, 7), holidays)
        # 8. Armed Forces Day
        self.assertIn(date(2020, 9, 25), holidays)
        # 9. Peace And Reconciliation Day
        self.assertIn(date(2020, 10, 4), holidays)
        # 10. Christmas day
        self.assertIn(date(2020, 12, 25), holidays)

    def test_2020_new_years_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 1, 1)], "New year")

    def test_2020_heroes_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 2, 3)], "Mozambican Heroes' Day")

    def test_2020_women_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 4, 7)], "Mozambican Women's Day")

    def test_2020_good_friday_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 4, 10)], "Good Friday")

    def test_2020_labour_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 5, 1)], "Labour Day")

    def test_2020_independence_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 6, 25)], "Independence Day")

    def test_2020_victory_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 9, 7)], "Victory Day")

    def test_2020_armed_forces_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 9, 25)], "Armed Forces Day")

    def test_2020_peace_and_reconciliation_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 10, 4)], "Peace And Reconciliation Day")

    def test_2020_christmas_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 12, 25)], "Christmas Day")



from datetime import date
from unittest import TestCase

from . import GenericCalendarTest
from ..europe import (
    Germany, BadenWurttemberg, Bavaria, Berlin, Brandenburg, Bremen,
    Hamburg, Hesse, MecklenburgVorpommern, LowerSaxony, NorthRhineWestphalia,
    RhinelandPalatinate, Saarland, Saxony, SaxonyAnhalt, SchleswigHolstein,
    Thuringia
)


class ReformationFlagTest(TestCase):
    def test_flags_reformation_day(self):
        klasses = (
            BadenWurttemberg, Bavaria, Berlin, Brandenburg, Bremen, Hamburg,
            Hesse, MecklenburgVorpommern, LowerSaxony, NorthRhineWestphalia,
            RhinelandPalatinate, Saarland, Saxony, SaxonyAnhalt,
            SchleswigHolstein, Thuringia
        )
        for klass in klasses:
            if klass in (Bremen, Hamburg, LowerSaxony, SchleswigHolstein):
                self.assertTrue(klass.include_reformation_day_2018)
            else:
                self.assertFalse(klass.include_reformation_day_2018)


class GermanyTest(GenericCalendarTest):
    cal_class = Germany

    def test_year_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 1, 1), holidays)
        self.assertIn(date(2014, 4, 18), holidays)
        self.assertIn(date(2014, 4, 21), holidays)
        self.assertIn(date(2014, 5, 1), holidays)
        self.assertIn(date(2014, 5, 29), holidays)
        self.assertIn(date(2014, 6, 9), holidays)
        self.assertIn(date(2014, 10, 3), holidays)
        self.assertIn(date(2014, 12, 25), holidays)
        self.assertIn(date(2014, 12, 26), holidays)

    def test_year_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 1), holidays)
        self.assertIn(date(2015, 4, 3), holidays)
        self.assertIn(date(2015, 4, 6), holidays)
        self.assertIn(date(2015, 5, 1), holidays)
        self.assertIn(date(2015, 5, 14), holidays)
        self.assertIn(date(2015, 5, 25), holidays)
        self.assertIn(date(2015, 10, 3), holidays)
        self.assertIn(date(2015, 12, 25), holidays)
        self.assertIn(date(2015, 12, 26), holidays)

    def test_year_reformation_day(self):
        # Reformation Day is included each year in:
        # Brandenburg, MecklenburgVorpommern, Saxony, SaxonyAnhalt & Thuringia
        holidays = self.cal.holidays_set(2015)
        if self.cal.include_reformation_day(2015):
            self.assertIn(date(2015, 10, 31), holidays)
        else:
            self.assertNotIn(date(2015, 10, 31), holidays)

        # But in the year 2017, it's included for the whole country
        holidays = self.cal.holidays_set(2017)
        self.assertTrue(self.cal.include_reformation_day(2017))
        self.assertIn(date(2017, 10, 31), holidays)

        # In 2018, Four more states have added it:
        # Bremen, Hamburg, Lower Saxony and Schleswig-Holstein
        holidays = self.cal.holidays_set(2018)
        if self.cal.include_reformation_day(2018):
            self.assertIn(date(2018, 10, 31), holidays)
        else:
            self.assertNotIn(date(2018, 10, 31), holidays)


class BadenWurttembergTest(GermanyTest):
    cal_class = BadenWurttemberg

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 1, 6), holidays)
        self.assertIn(date(2014, 6, 19), holidays)
        self.assertIn(date(2014, 11, 1), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 6), holidays)
        self.assertIn(date(2015, 6, 4), holidays)
        self.assertIn(date(2015, 11, 1), holidays)


class BavariaTest(GenericCalendarTest):
    cal_class = Bavaria

    def test_year_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 1, 1), holidays)
        self.assertIn(date(2014, 1, 6), holidays)
        self.assertIn(date(2014, 4, 18), holidays)
        self.assertIn(date(2014, 4, 21), holidays)
        self.assertIn(date(2014, 5, 1), holidays)
        self.assertIn(date(2014, 5, 29), holidays)
        self.assertIn(date(2014, 6, 9), holidays)
        self.assertIn(date(2014, 6, 19), holidays)
        self.assertIn(date(2014, 8, 15), holidays)
        self.assertIn(date(2014, 10, 3), holidays)
        self.assertIn(date(2014, 11, 1), holidays)
        self.assertIn(date(2014, 12, 25), holidays)
        self.assertIn(date(2014, 12, 26), holidays)

    def test_year_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 1), holidays)
        self.assertIn(date(2015, 1, 6), holidays)
        self.assertIn(date(2015, 4, 3), holidays)
        self.assertIn(date(2015, 4, 6), holidays)
        self.assertIn(date(2015, 5, 1), holidays)
        self.assertIn(date(2015, 5, 14), holidays)
        self.assertIn(date(2015, 5, 25), holidays)
        self.assertIn(date(2015, 6, 4), holidays)
        self.assertIn(date(2015, 8, 15), holidays)
        self.assertIn(date(2015, 10, 3), holidays)
        self.assertIn(date(2015, 11, 1), holidays)
        self.assertIn(date(2015, 12, 25), holidays)
        self.assertIn(date(2015, 12, 26), holidays)


class BerlinTest(GermanyTest):
    cal_class = Berlin

    def test_extra_2018(self):
        holidays = self.cal.holidays_set(2018)
        self.assertNotIn(date(2018, 3, 8), holidays)
        self.assertNotIn(date(2018, 5, 8), holidays)

    def test_extra_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 3, 8), holidays)
        self.assertNotIn(date(2019, 5, 8), holidays)

    def test_extra_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 3, 8), holidays)
        self.assertIn(date(2020, 5, 8), holidays)

    def test_extra_2021(self):
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 3, 8), holidays)
        self.assertNotIn(date(2021, 5, 8), holidays)


class BrandenburgTest(GenericCalendarTest):
    cal_class = Brandenburg

    def test_year_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 1, 1), holidays)
        self.assertIn(date(2014, 4, 18), holidays)
        self.assertIn(date(2014, 4, 20), holidays)
        self.assertIn(date(2014, 4, 21), holidays)
        self.assertIn(date(2014, 5, 1), holidays)
        self.assertIn(date(2014, 5, 29), holidays)
        self.assertIn(date(2014, 6, 9), holidays)
        self.assertIn(date(2014, 10, 3), holidays)
        self.assertIn(date(2014, 10, 31), holidays)
        self.assertIn(date(2014, 12, 25), holidays)
        self.assertIn(date(2014, 12, 26), holidays)

    def test_year_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 1), holidays)
        self.assertIn(date(2015, 4, 3), holidays)
        self.assertIn(date(2015, 4, 5), holidays)
        self.assertIn(date(2015, 4, 6), holidays)
        self.assertIn(date(2015, 5, 1), holidays)
        self.assertIn(date(2015, 5, 14), holidays)
        self.assertIn(date(2015, 5, 25), holidays)
        self.assertIn(date(2015, 10, 3), holidays)
        self.assertIn(date(2015, 10, 31), holidays)
        self.assertIn(date(2015, 12, 25), holidays)
        self.assertIn(date(2015, 12, 26), holidays)


class BremenTest(GermanyTest):
    cal_class = Bremen


class HamburgTest(GermanyTest):
    cal_class = Hamburg


class HesseTest(GermanyTest):
    cal_class = Hesse

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 6, 19), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 6, 4), holidays)


class MecklenburgVorpommernTest(GermanyTest):
    cal_class = MecklenburgVorpommern

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 10, 31), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 10, 31), holidays)


class LowerSaxonyTest(GermanyTest):
    cal_class = LowerSaxony


class NorthRhineWestphaliaTest(GermanyTest):
    cal_class = NorthRhineWestphalia

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 6, 19), holidays)
        self.assertIn(date(2014, 11, 1), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 6, 4), holidays)
        self.assertIn(date(2015, 11, 1), holidays)


class RhinelandPalatinateTest(GermanyTest):
    cal_class = RhinelandPalatinate

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 6, 19), holidays)
        self.assertIn(date(2014, 11, 1), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 6, 4), holidays)
        self.assertIn(date(2015, 11, 1), holidays)


class SaarlandTest(GermanyTest):
    cal_class = Saarland

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 6, 19), holidays)
        self.assertIn(date(2014, 8, 15), holidays)
        self.assertIn(date(2014, 11, 1), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 6, 4), holidays)
        self.assertIn(date(2015, 8, 15), holidays)
        self.assertIn(date(2015, 11, 1), holidays)


class SaxonyTest(GermanyTest):
    cal_class = Saxony

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 11, 19), holidays)
        self.assertIn(date(2014, 10, 31), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 11, 18), holidays)
        self.assertIn(date(2015, 10, 31), holidays)

    def test_extra_2016(self):
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 11, 16), holidays)
        self.assertNotIn(date(2016, 11, 23), holidays)
        self.assertIn(date(2016, 10, 31), holidays)


class SaxonyAnhaltTest(GermanyTest):
    cal_class = SaxonyAnhalt

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 10, 31), holidays)
        self.assertIn(date(2014, 1, 6), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 10, 31), holidays)
        self.assertIn(date(2015, 1, 6), holidays)


class SchleswigHolsteinTest(GermanyTest):
    cal_class = SchleswigHolstein


class ThuringiaTest(GermanyTest):
    cal_class = Thuringia

    def test_extra_2014(self):
        holidays = self.cal.holidays_set(2014)
        self.assertIn(date(2014, 10, 31), holidays)

    def test_extra_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 10, 31), holidays)



from datetime import date, timedelta
from ..core import WesternCalendar, SUN, ISO_SAT
from ..registry_tools import iso_register


@iso_register('NL')
class Netherlands(WesternCalendar):
    'Netherlands'

    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_ascension = True
    include_whit_sunday = True
    include_whit_monday = True
    include_boxing_day = True

    observance_shift = None

    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (5, 5, "Liberation Day"),
    )

    def __init__(self, include_carnival=False):
        self.include_carnival = include_carnival
        super().__init__()

    def get_king_queen_day(self, year):
        """27 April unless this is a Sunday in which case it is the 26th

        Before 2013 it was called Queensday, falling on
        30 April, unless this is a Sunday in which case it is the 29th.
        """
        if year > 2013:
            king_day = date(year, 4, 27)
            if king_day.weekday() != SUN:
                return king_day, "King's day"
            return king_day - timedelta(days=1), "King's day"
        else:
            queen_day = date(year, 4, 30)
            if queen_day.weekday() != SUN:
                return queen_day, "Queen's day"
            return queen_day - timedelta(days=1), "Queen's day"

    def get_carnival_days(self, year):
        """Carnival starts 7 weeks before Easter Sunday and lasts 3 days."""
        n_days = 3
        start = self.get_easter_sunday(year) - timedelta(weeks=7)

        return [
            (
                start + timedelta(days=i), "Carnival"
            ) for i in range(n_days)
        ]

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(self.get_king_queen_day(year))
        if self.include_carnival:
            days.extend(self.get_carnival_days(year))
        return days


FALL_HOLIDAYS_EARLY_REGIONS = {
    2025: ["south"],
    2024: ["south"],
    2023: ["middle", "south"],
    2022: ["north"],
    2021: ["north", "middle"],
    2020: ["north"],
    2019: ["south"],
    2018: ["south"],
    2017: ["middle", "south"],
    2016: ["north", "middle"],
}

SPRING_HOLIDAYS_EARLY_REGIONS = {
    2025: ["north"],
    2024: ["south"],
    2023: ["south"],
    2022: ["north"],
    2021: ["south"],
    2020: ["north"],
    2019: ["north"],
    2018: ["south"],
    2017: ["north"],
    2016: ["middle", "south"],
}

SUMMER_HOLIDAYS_EARLY_REGIONS = {
    2025: ["south"],
    2024: ["south"],
    2023: ["middle"],
    2022: ["middle"],
    2021: ["north"],
    2020: ["north"],
    2019: ["south"],
    2018: ["south"],
    2017: ["middle"],
    2016: ["middle"],
}

SUMMER_HOLIDAYS_LATE_REGIONS = {
    2025: ["middle"],
    2024: ["north"],
    2023: ["north"],
    2022: ["south"],
    2021: ["south"],
    2020: ["middle"],
    2019: ["middle"],
    2018: ["north"],
    2017: ["north"],
    2016: ["south"],
}


class NetherlandsWithSchoolHolidays(Netherlands):
    """Netherlands with school holidays (2016 to 2025).

    Data source and regulating body:
    https://www.rijksoverheid.nl/onderwerpen/schoolvakanties/overzicht-schoolvakanties-per-schooljaar
    """

    def __init__(self, region, carnival_instead_of_spring=False, **kwargs):
        """ Set up a calendar incl. school holidays for a specific region

        :param region: either "north", "middle" or "south"
        """
        if region not in ("north", "middle", "south"):
            raise ValueError("Set region to 'north', 'middle' or 'south'.")
        self.region = region
        self.carnival_instead_of_spring = carnival_instead_of_spring
        if carnival_instead_of_spring and "include_carnival" not in kwargs:
            kwargs["include_carnival"] = True
        super().__init__(**kwargs)

    def get_fall_holidays(self, year):
        """
        Return Fall holidays.

        They start at week 43 or 44 and last for 9 days
        """
        if year not in FALL_HOLIDAYS_EARLY_REGIONS:
            raise NotImplementedError(f"Unknown fall holidays for {year}.")

        n_days = 9
        week = 43

        # Exceptional years
        if year == 2024:
            week = 44

        # Holiday starts on the preceding Saturday
        start = self.get_iso_week_date(year, week - 1, ISO_SAT)

        # Some regions have their fall holiday 1 week earlier
        if self.region in FALL_HOLIDAYS_EARLY_REGIONS[year]:
            start = start - timedelta(weeks=1)

        return [
            (start + timedelta(days=i), "Fall holiday") for i in range(n_days)
        ]

    def get_christmas_holidays(self, year):
        """
        Return Christmas holidays

        Christmas holidays run partially in December and partially in January
        (spillover from previous year).
        """
        return (
            self._get_christmas_holidays_december(year) +
            self._get_christmas_holidays_january(year)
        )

    def _get_christmas_holidays_december(self, year):
        # 27 December is always in a full week of holidays
        week = date(year, 12, 27).isocalendar()[1]

        # Holiday starts on the preceding Saturday
        start = self.get_iso_week_date(year, week - 1, ISO_SAT)
        return [
            (start + timedelta(days=i), "Christmas holiday")
            for i in range((date(year, 12, 31) - start).days + 1)
        ]

    def _get_christmas_holidays_january(self, year):
        # 27 December is always in a full week of holidays
        week = date(year - 1, 12, 27).isocalendar()[1]

        # Holiday ends 15 days after the preceding Saturday
        # Saturday of the previous week (previous year!)
        start = self.get_iso_week_date(year - 1, week - 1, ISO_SAT)
        end = start + timedelta(days=15)

        return [
            (date(year, 1, 1) + timedelta(days=i), "Christmas holiday")
            for i in range((end - date(year, 1, 1)).days + 1)
        ]

    def get_spring_holidays(self, year):
        """
        Return the Spring holidays

        They start at week 8 or 9 and last for 9 days.
        """
        if year not in SPRING_HOLIDAYS_EARLY_REGIONS:
            raise NotImplementedError(f"Unknown spring holidays for {year}.")

        n_days = 9
        week = 9

        # Exceptional years
        if year in [2024, 2021]:
            week = 8

        # Holiday starts on the preceding Saturday
        start = self.get_iso_week_date(year, week - 1, ISO_SAT)

        # Some regions have their spring holiday 1 week earlier
        if self.region in SPRING_HOLIDAYS_EARLY_REGIONS[year]:
            start = start - timedelta(weeks=1)

        return [
            (start + timedelta(days=i), "Spring holiday")
            for i in range(n_days)
        ]

    def get_carnival_holidays(self, year):
        """
        Return Carnival holidays

        Carnival holidays start 7 weeks and 1 day before Easter Sunday
        and last 9 days.
        """
        n_days = 9
        start = self.get_easter_sunday(year) - timedelta(weeks=7, days=1)

        return [
            (start + timedelta(days=i), "Carnival holiday")
            for i in range(n_days)
        ]

    def get_may_holidays(self, year):
        """
        Return May holidays

        They start at week 18 (or 17) and last for 18 days
        """
        n_days = 9
        week = 18

        # Exceptional years
        if year == 2017:
            week = 17

        # Holiday starts on the preceding Saturday
        start = self.get_iso_week_date(year, week - 1, ISO_SAT)

        return [
            (start + timedelta(days=i), "May holiday") for i in range(n_days)
        ]

    def get_summer_holidays(self, year):
        """
        Return the summer holidays as a list
        """

        if year not in SUMMER_HOLIDAYS_EARLY_REGIONS or \
                year not in SUMMER_HOLIDAYS_LATE_REGIONS:
            raise NotImplementedError(f"Unknown summer holidays for {year}.")

        n_days = 44
        week = 29

        # Holiday starts on the preceding Saturday
        start = self.get_iso_week_date(year, week - 1, ISO_SAT)

        # Some regions have their summer holiday 1 week earlier
        if self.region in SUMMER_HOLIDAYS_EARLY_REGIONS[year]:
            start = start - timedelta(weeks=1)

        if self.region in SUMMER_HOLIDAYS_LATE_REGIONS[year]:
            start = start + timedelta(weeks=1)

        return [
            (start + timedelta(days=i), "Summer holiday")
            for i in range(n_days)
        ]

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.extend(self.get_fall_holidays(year))
        days.extend(self.get_christmas_holidays(year))
        if self.carnival_instead_of_spring:
            days.extend(self.get_carnival_holidays(year))
        else:
            days.extend(self.get_spring_holidays(year))
        days.extend(self.get_may_holidays(year))
        days.extend(self.get_summer_holidays(year))
        return days



from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('MC')
class Monaco(WesternCalendar):
    'Monaco'

    # Christian holidays
    include_easter_monday = True
    include_ascension = True
    include_whit_monday = True
    include_all_saints = True
    include_assumption = True
    include_corpus_christi = True
    include_immaculate_conception = True

    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (1, 27, "Saint Dévote's Day"),
        (11, 19, "H.S.H. the Sovereign Prince's Day"),
    )
